{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Recipe 6: Analyzing a Non-Parametric Reform\n",
    "\n",
    "This is an advanced recipe that should be followed only after mastering the basic recipe.\n",
    "This recipe shows how to customize the Calculator class so that it can analyze a tax reform\n",
    "that cannot be characterized using existing policy parameters (in this case a pseudo Cost-of-Living Refund reform).\n",
    "It uses, in a more extensive way, the object-oriented programming inheritance technique introduced in Redefining Expanded Income,\n",
    "so you might want to read that recipe first.\n",
    "\n",
    "Recipe that illustrates how to customize the taxcalc Calculator class so that\n",
    "it can analyze a non-parametric reform (that is, a reform that cannot be\n",
    "represented using existing taxcalc Policy parameters.\n",
    "\n",
    "The technique for doing this customization is standard in object-oriented\n",
    "programming: a child class is derived from a parent class and then customized.\n",
    "The derived child class inherits all the data and methods of the parent class,\n",
    "but can be customized by adding new data and methods or by overriding inherited\n",
    "methods.\n",
    "\n",
    "The reform used to illustrate this programming technique is somewhat like\n",
    "the Cost-of-Living Refund, a refundable credit that was being discussed in\n",
    "tax policy circles during 2019 as a replacement for the EITC.  But the\n",
    "reform analyzed here is not exactly like the Cost-of-Living Refund, so we\n",
    "call it a pseudo cost-of-living refund to emphasize that it is not meant\n",
    "to be an accurate representation of the Cost-of-Living Refund proposal."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "tags": [
     "remove-cell"
    ]
   },
   "outputs": [],
   "source": [
    "# Install taxcalc if not already installed\n",
    "import sys\n",
    "if 'taxcalc' not in sys.modules:\n",
    "    %pip install taxcalc &> /dev/null  # dev/null suppresses output"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "hide-output": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "YEAR  UNITS   ITAX1   ITAX2  EITC1  EITC2  COLR1  COLR2\n",
      "2019  201.9  1259.0  1259.0   63.3   63.3    0.0    0.0\n",
      "2020  204.5  764.2  473.6   65.5    0.0    0.0  356.1\n",
      "2021  207.1  949.3  694.1   76.4    0.0    0.0  331.6\n",
      "2022  209.7  1674.6  1386.5   60.7    0.0    0.0  348.8\n"
     ]
    }
   ],
   "source": [
    "import numpy as np\n",
    "import taxcalc as tc\n",
    "\n",
    "\n",
    "class Calculator(tc.Calculator):\n",
    "    \"\"\"\n",
    "    Customized Calculator class that inherits all tc.Calculator data and\n",
    "    methods, adding or overriding some to get the desired customization.\n",
    "    \"\"\"\n",
    "    def __init__(self, policy=None, records=None, verbose=False,\n",
    "                 sync_years=True, consumption=None,\n",
    "                 # above are the tc.Calculator constructor arguments\n",
    "                 # below is new constructor argument used in customization\n",
    "                 colr_active=False):\n",
    "        # call the parent class constructor\n",
    "        super().__init__(policy=policy, records=records,\n",
    "                         verbose=verbose, sync_years=sync_years,\n",
    "                         consumption=consumption)\n",
    "        # remember whether pseudo_COLR policy is active or not\n",
    "        self.colr_active = colr_active\n",
    "        # declare colr_param dictionary that will contain pseudo COLR policy\n",
    "        self.colr_param = dict()\n",
    "\n",
    "    def specify_pseudo_colr_policy(self):\n",
    "        \"\"\"\n",
    "        Specify policy parameters for the COLR policy in the current_year.\n",
    "        See use of these parameters below in the pseudo_colr_amount method.\n",
    "        \"\"\"\n",
    "        # reform implementation year\n",
    "        reform_year = 2020\n",
    "        # specify dictionary of parameter names and values for reform_year\n",
    "        self.colr_param = {\n",
    "            # credit phase-in rate on earnings\n",
    "            'COLR_rt': 1.0,\n",
    "            # ceiling on refundable credit varies by filing-unit type, MARS\n",
    "            'COLR_c': np.array([4000, 8000, 4000, 4000, 4000],\n",
    "                               dtype=np.float64),\n",
    "            # credit phase-out start AGI level varies by filing-unit type, MARS\n",
    "            'COLR_ps': np.array([30000, 50000, 30000, 30000, 30000],\n",
    "                                dtype=np.float64),\n",
    "            # credit phase-out rate per dollar of AGI above COLR_ps level\n",
    "            'COLR_prt': 0.2\n",
    "        }\n",
    "        # set pseudo COLR parameter values for current year\n",
    "        this_year = self.current_year\n",
    "        if self.colr_active and this_year >= reform_year:\n",
    "            # set inflation-indexed values of COLR_c and COLR_ps for this year\n",
    "            irates = self.__policy.inflation_rates()\n",
    "            syr = tc.Policy.JSON_START_YEAR\n",
    "            for name in ['COLR_c', 'COLR_ps']:\n",
    "                value = self.colr_param[name]\n",
    "                for year in range(reform_year, this_year):\n",
    "                    value *= (1.0 + irates[year - syr])\n",
    "                self.colr_param[name] = np.round(value, 2)  # to nearest penny\n",
    "        else:  # if policy not active or if this year is before the reform year\n",
    "            # set ceiling to zero\n",
    "            self.colr_param['COLR_c'] = np.array([0.0, 0.0, 0.0, 0.0, 0.0],\n",
    "                                                 dtype=np.float64)\n",
    "        tracing = False  # set to True to see parameter values for this year\n",
    "        if tracing:\n",
    "            for name in self.colr_param:\n",
    "                print('> {} {} {}'.format(\n",
    "                    this_year, name, self.colr_param[name]\n",
    "                ))\n",
    "\n",
    "    def pseudo_colr_amount(self):\n",
    "        \"\"\"\n",
    "        Calculate pseudo Cost-of-Living Refund amount.\n",
    "        Note this is simply meant to illustrate a Python programming technique;\n",
    "        this function does NOT calculate an exact Cost-of-Living Refund amount.\n",
    "        See setting of parameters above in specify_pseudo_COLR_policy method.\n",
    "        \"\"\"\n",
    "        recs = self.__records\n",
    "        # create MARS-specific policy parameter arrays\n",
    "        mars_indicators = [recs.MARS == 1, recs.MARS == 2, recs.MARS == 3,\n",
    "                           recs.MARS == 4, recs.MARS == 5]\n",
    "        colr_c = np.select(mars_indicators, self.colr_param['COLR_c'])\n",
    "        colr_ps = np.select(mars_indicators, self.colr_param['COLR_ps'])\n",
    "        colr_rt = self.colr_param['COLR_rt']\n",
    "        colr_prt = self.colr_param['COLR_prt']\n",
    "        # compute colr_amt\n",
    "        amt_pre_phaseout = np.minimum(recs.e00200 * colr_rt, colr_c)\n",
    "        phaseout = np.maximum((recs.c00100 - colr_ps) * colr_prt, 0.)\n",
    "        colr_amt = np.maximum(amt_pre_phaseout - phaseout, 0.)\n",
    "        setattr(recs, 'colr_amount', colr_amt)\n",
    "        # reduce income and combined taxes because COLR is a refundable credit\n",
    "        recs.iitax -= colr_amt\n",
    "        recs.combined -= colr_amt\n",
    "        # delete local arrays used only in this method\n",
    "        del mars_indicators\n",
    "        del colr_c\n",
    "        del colr_ps\n",
    "        del amt_pre_phaseout\n",
    "        del phaseout\n",
    "        del colr_amt\n",
    "\n",
    "    def calc_all(self, zero_out_calc_vars=False):\n",
    "        \"\"\"\n",
    "        Call all tax-calculation functions for the current_year.\n",
    "        \"\"\"\n",
    "        tc.BenefitPrograms(self)\n",
    "        self._calc_one_year(zero_out_calc_vars)\n",
    "        tc.BenefitSurtax(self)\n",
    "        tc.BenefitLimitation(self)\n",
    "        tc.FairShareTax(self.__policy, self.__records)\n",
    "        tc.LumpSumTax(self.__policy, self.__records)\n",
    "        # specify new method to set pseudo COLR policy parameters\n",
    "        self.specify_pseudo_colr_policy()  # (see above)\n",
    "        # call new method to calculate pseudo COLR amount\n",
    "        self.pseudo_colr_amount()  # (see above)\n",
    "        tc.ExpandIncome(self.__policy, self.__records)\n",
    "        tc.AfterTaxIncome(self.__policy, self.__records)\n",
    "\n",
    "# end of customized Calculator class definition\n",
    "\n",
    "\n",
    "# top-level logic of program that uses customized Calculator class\n",
    "\n",
    "policy1 = tc.Policy()  # baseline policy is current-law policy\n",
    "policy2 = tc.Policy()  # parametric reform, reformC.json, eliminates EITC\n",
    "\n",
    "# TODO: Move to the web so this can be done standalone.\n",
    "policy2.implement_reform(tc.Policy.read_json_reform('github://PSLmodels:Tax-Calculator@master/docs/recipes/_static/reformC.json'))\n",
    "\n",
    "# specify customized Calculator objects for baseline and reform:\n",
    "#   baseline calc1 uses policy1 (current-law) and colr_active=False\n",
    "#   reform calc2 uses policy2 (no EITC) and colr_active=True\n",
    "cps_records = tc.Records.cps_constructor()\n",
    "calc1 = Calculator(policy=policy1, records=cps_records, colr_active=False)\n",
    "calc2 = Calculator(policy=policy2, records=cps_records, colr_active=True)\n",
    "\n",
    "# calculate tax liabilities for years around the reform year\n",
    "CYR_FIRST = 2019\n",
    "CYR_LAST = 2022\n",
    "for cyr in range(CYR_FIRST, CYR_LAST + 1):\n",
    "    # advance to and calculate for specified cyr\n",
    "    calc1.advance_to_year(cyr)\n",
    "    calc1.calc_all()\n",
    "    calc2.advance_to_year(cyr)\n",
    "    calc2.calc_all()\n",
    "    # tabulate weighted amounts\n",
    "    funits = calc1.total_weight()\n",
    "    itax1 = calc1.weighted_total('iitax')\n",
    "    itax2 = calc2.weighted_total('iitax')\n",
    "    eitc1 = calc1.weighted_total('eitc')\n",
    "    eitc2 = calc2.weighted_total('eitc')\n",
    "    colr1 = calc1.weighted_total('colr_amount')\n",
    "    colr2 = calc2.weighted_total('colr_amount')\n",
    "    # print weighted amounts for cyr\n",
    "    if cyr == CYR_FIRST:\n",
    "        print('YEAR  UNITS   ITAX1   ITAX2  EITC1  EITC2  COLR1  COLR2')\n",
    "    line = '{}  {:.1f}  {:.1f}  {:.1f}  {:5.1f}  {:5.1f}  {:5.1f}  {:5.1f}'\n",
    "    print(line.format(cyr, funits * 1e-6,\n",
    "                      itax1 * 1e-9, itax2 * 1e-9,\n",
    "                      eitc1 * 1e-9, eitc2 * 1e-9,\n",
    "                      colr1 * 1e-9, colr2 * 1e-9))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "celltoolbar": "Tags",
  "date": 1592702193.422727,
  "download_nb": false,
  "filename": "recipe06.rst",
  "filename_with_path": "recipe06",
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.6"
  },
  "title": "Recipe 6: Analyzing a Non-Parametric Reform"
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
